from collections.abc import MutableSequence
from typing import Union, Iterable, Tuple, Sequence, List

import numpy as np

from ...helper import typename
from ...logging import default_logger
from ...enums import EmbeddingClsType

try:
    # when protobuf using Cpp backend
    from google.protobuf.pyext._message import (
        RepeatedCompositeContainer as RepeatedContainer,
    )
except:
    # when protobuf using Python backend
    from google.protobuf.internal.containers import (
        RepeatedCompositeFieldContainer as RepeatedContainer,
    )

from ...proto.jina_pb2 import DocumentProto
from .traversable import TraversableSequence

__all__ = ['DocumentArray']

if False:
    from ..document import Document
    from scipy.sparse import coo_matrix

    # fix type-hint complain for sphinx and flake
    from typing import TypeVar
    import scipy
    import tensorflow as tf
    import torch

    EmbeddingType = TypeVar(
        'EncodingType',
        np.ndarray,
        scipy.sparse.csr_matrix,
        scipy.sparse.coo_matrix,
        scipy.sparse.bsr_matrix,
        scipy.sparse.csc_matrix,
        torch.sparse_coo_tensor,
        tf.SparseTensor,
    )

    SparseEmbeddingType = TypeVar(
        'SparseEmbeddingType',
        np.ndarray,
        scipy.sparse.csr_matrix,
        scipy.sparse.coo_matrix,
        scipy.sparse.bsr_matrix,
        scipy.sparse.csc_matrix,
        torch.sparse_coo_tensor,
        tf.SparseTensor,
    )


class DocumentArray(TraversableSequence, MutableSequence):
    """
    :class:`DocumentArray` is a mutable sequence of :class:`Document`.
    It gives an efficient view of an array of Documents.
    One can iterate over it like a generator but ALSO modify it, count it,
    get item, or union two 'DocumentArray's using the '+' and '+=' operators.

    :param docs_proto: A list of :class:`Document`
    :type docs_proto: Union['RepeatedContainer', Sequence['Document']]
    """

    def __init__(self, docs_proto: Union['RepeatedContainer', Sequence['Document']]):
        super().__init__()
        self._docs_proto = docs_proto
        self._docs_map = {}

    def insert(self, index: int, doc: 'Document') -> None:
        """
        Insert :param:`doc.proto` at :param:`index` into the array of `:class:`DocumentArray` .
        :param index: Position of the insertion.
        :param doc: The doc to be inserted.
        """
        self._docs_proto.insert(index, doc.proto)

    def __setitem__(self, key, value: 'Document'):
        if isinstance(key, int):
            self._docs_proto[key].CopyFrom(value)
        elif isinstance(key, str):
            self._docs_map[key].CopyFrom(value)
        else:
            raise IndexError(f'do not support this index {key}')

    def __delitem__(self, index):
        del self._docs_proto[index]

    def __len__(self):
        return len(self._docs_proto)

    def __iter__(self) -> 'Document':
        from ..document import Document

        for d in self._docs_proto:
            yield Document(d)

    def __getitem__(self, item):
        from ..document import Document

        if isinstance(item, int):
            return Document(self._docs_proto[item])
        elif isinstance(item, str):
            return Document(self._docs_map[item])
        elif isinstance(item, slice):
            return DocumentArray(self._docs_proto[item])
        else:
            raise IndexError(f'do not support this index {item}')

    def __add__(self, other: 'DocumentArray'):
        v = DocumentArray([])
        for doc in self:
            v.add(doc)
        for doc in other:
            v.add(doc)
        return v

    def __iadd__(self, other: 'DocumentArray'):
        for doc in other:
            self.add(doc)
        return self

    def append(self, doc: 'Document') -> 'Document':
        """
        Append :param:`doc` in :class:`DocumentArray`.
        :param doc: The doc needs to be appended.
        :return: Appended internal list.
        """
        return self._docs_proto.append(doc.proto)

    def add(self, doc: 'Document') -> 'Document':
        """Shortcut to :meth:`append`, do not override this method.
        :param doc: the document to add to the array
        :return: Appended internal list.
        """
        return self.append(doc)

    def extend(self, iterable: Iterable['Document']) -> None:
        """
        Extend the :class:`DocumentArray` by appending all the items from the iterable.
        :param iterable: the iterable of Documents to extend this array with
        """
        for doc in iterable:
            self.append(doc)

    def clear(self):
        """Clear the data of :class:`DocumentArray`"""
        del self._docs_proto[:]

    def reverse(self):
        """In-place reverse the sequence."""
        if isinstance(self._docs_proto, RepeatedContainer):
            size = len(self._docs_proto)
            hi_idx = size - 1
            for i in range(int(size / 2)):
                tmp = DocumentProto()
                tmp.CopyFrom(self._docs_proto[hi_idx])
                self._docs_proto[hi_idx].CopyFrom(self._docs_proto[i])
                self._docs_proto[i].CopyFrom(tmp)
                hi_idx -= 1
        elif isinstance(self._docs_proto, list):
            self._docs_proto.reverse()

    def build(self):
        """Build a doc_id to doc mapping so one can later index a Document using doc_id as string key."""
        self._docs_map = {d.id: d for d in self._docs_proto}

    def sort(self, *args, **kwargs):
        """
        Sort the items of the :class:`DocumentArray` in place.


        :param args: variable list of arguments to pass to the sorting underlying function
        :param kwargs: keyword arguments to pass to the sorting underlying function
        """
        self._docs_proto.sort(*args, **kwargs)

    @property
    def all_embeddings(self) -> Tuple['np.ndarray', 'DocumentArray']:
        """Return all embeddings from every document in this array as a ndarray
        :return: The corresponding documents in a :class:`DocumentArray`,
                and the documents have no embedding in a :class:`DocumentArray`.
        :rtype: A tuple of embedding in :class:`np.ndarray`
        """
        return self.extract_docs('embedding', stack_contents=True)

    def get_all_sparse_embeddings(
        self, embedding_cls_type: EmbeddingClsType
    ) -> Tuple['SparseEmbeddingType', 'DocumentArray']:
        """Return all embeddings from every document in this array as a sparse array

        :param embedding_cls_type: Type of sparse matrix backend, e.g. `scipy`, `torch` or `tf`.
        :return: The corresponding documents in a :class:`DocumentArray`,
            and the documents have no embedding in a :class:`DocumentArray`.
        :rtype: A tuple of embedding and DocumentArray as sparse arrays
        """

        def stack_embeddings(embeddings):
            if embedding_cls_type.is_scipy:
                import scipy

                return scipy.sparse.vstack(embeddings)
            elif embedding_cls_type.is_torch:
                import torch

                return torch.vstack(embeddings)
            elif embedding_cls_type.is_tf:
                return embeddings
            else:
                raise ValueError(
                    f'Trying to stack sparse embeddings with embedding_cls_type {embedding_cls_type} failed'
                )

        def get_sparse_ndarray_type_kwargs():
            if embedding_cls_type.is_scipy:
                from jina.types.ndarray.sparse.scipy import SparseNdArray

                if not embedding_cls_type.is_scipy_stackable not in ['coo', 'csr']:
                    default_logger.warning(
                        f'found `{embedding_cls_type.name}` matrix, recommend to use `coo` or `csr` type.'
                    )
                return SparseNdArray, {'sp_format': embedding_cls_type.scipy_cls_type}
            elif embedding_cls_type.is_torch:
                from jina.types.ndarray.sparse.pytorch import SparseNdArray

                return SparseNdArray, {}
            elif embedding_cls_type.is_tf:
                from jina.types.ndarray.sparse.tensorflow import SparseNdArray

                return SparseNdArray, {}
            else:
                raise ValueError(
                    f'Trying to get sparse embeddings with embedding_cls_type {embedding_cls_type} failed'
                )

        embeddings = []
        docs_pts = []
        bad_docs = []
        sparse_ndarray_type, sparse_kwargs = get_sparse_ndarray_type_kwargs()
        for doc in self:
            embedding = doc.get_sparse_embedding(
                sparse_ndarray_cls_type=sparse_ndarray_type, **sparse_kwargs
            )
            if embedding is None:
                bad_docs.append(doc)
                continue
            embeddings.append(embedding)
            docs_pts.append(doc)

        if bad_docs:
            default_logger.warning(
                f'found {len(bad_docs)} docs at granularity {bad_docs[0].granularity} are missing sparse_embedding'
            )

        return stack_embeddings(embeddings), docs_pts

    @property
    def all_contents(self) -> Tuple['np.ndarray', 'DocumentArray']:
        """Return all embeddings from every document in this array as a ndarray
        :return: The corresponding documents in a :class:`DocumentArray`,
                and the documents have no contents in a :class:`DocumentArray`.
        :rtype: A tuple of embedding in :class:`np.ndarray`
        """
        # stack true for backward compatibility, but will not work if content is blob of different shapes
        return self.extract_docs('content', stack_contents=True)

    def extract_docs(
        self, *fields: str, stack_contents: Union[bool, List[bool]] = False
    ) -> Tuple[Union['np.ndarray', List['np.ndarray']], 'DocumentArray']:
        """Return in batches all the values of the fields
        :param fields: Variable length argument with the name of the fields to extract
        :param stack_contents: boolean flag indicating if output arrays should be stacked with `np.stack`
        :return: Returns an :class:`np.ndarray` or an array of :class:`np.ndarray` with the batches for these fields
        """

        list_of_contents_output = len(fields) > 1
        contents = [[] for _ in fields if list_of_contents_output]
        docs_pts = []
        bad_docs = []

        if list_of_contents_output:
            for doc in self:
                content = doc.get_attrs_values(*fields)
                if content is None:
                    bad_docs.append(doc)
                    continue
                for i, c in enumerate(content):
                    contents[i].append(c)
                docs_pts.append(doc)
            for idx, c in enumerate(contents):
                if not c:
                    continue
                if (
                    isinstance(stack_contents, bool)
                    and stack_contents
                    and not isinstance(c[0], bytes)
                ) or (
                    isinstance(stack_contents, list)
                    and stack_contents[idx]
                    and not isinstance(c[0], bytes)
                ):
                    contents[idx] = np.stack(c)
        else:
            for doc in self:
                content = doc.get_attrs_values(*fields)[0]
                if content is None:
                    bad_docs.append(doc)
                    continue
                contents.append(content)
                docs_pts.append(doc)

            if not contents:
                contents = None
            elif (
                isinstance(stack_contents, bool)
                and stack_contents
                and not isinstance(contents[0], bytes)
            ) or (
                isinstance(stack_contents, list)
                and stack_contents[0]
                and not isinstance(contents[0], bytes)
            ):
                contents = np.stack(contents)

        if bad_docs:
            default_logger.warning(
                f'found {len(bad_docs)} docs at granularity {bad_docs[0].granularity} are missing one of the '
                f'following fields: {fields} '
            )

        if not docs_pts:
            default_logger.warning('no documents are extracted')

        return contents, DocumentArray(docs_pts)

    def __bool__(self):
        """To simulate ```l = []; if l: ...```
        :return: returns true if the length of the array is larger than 0
        """
        return len(self) > 0

    def new(self) -> 'Document':
        """Create a new empty document appended to the end of the array.
        :return: a new Document appended to the internal list
        """
        from ..document import Document

        return self.append(Document())

    def __str__(self):
        from ..document import Document

        content = ',\n'.join(str(Document(d)) for d in self._docs_proto[:3])
        if len(self._docs_proto) > 3:
            content += f'in total {len(self._docs_proto)} items'
        return content

    def __repr__(self):
        content = ' '.join(
            f'{k}={v}' for k, v in {'length': len(self._docs_proto)}.items()
        )
        content += f' at {id(self)}'
        content = content.strip()
        return f'<{typename(self)} {content}>'



from typing import Dict, List, Union

import numpy as np
from jina.executors.crafters import BaseCrafter

if False:
    import torch


class ImageTorchTransformation(BaseCrafter):
    """Apply torchvision transforms on image batches.

    This crafter creates a `Compose` transform using the list of transforms provided.
    The numpy input is converted to tensor before applying the transform. Hence `ToTensor()` is not required.

    Note: Not recommended to use inherently random transforms and
    make sure not to use transforms that do not support tensor inputs.

    Example:
        .. code-block:: python
            from jina.hub.crafters.image.ImageTorchTransformation import ImageTorchTransformation
            import numpy

            transforms = [
                {'CenterCrop': dict(size=(300))},
                {'Resize': dict(size=(244, 244))},
                'RandomVerticalFlip',
                {'Normalize': dict(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))},
            ]

            crafter = ImageTorchTransformation(transforms)
            crafted_imgs = crafter.craft(image_batch)

        The torch equivalent for this would be recursively apply `Compose` on each image.

        .. code-block:: python
            import torchvision.transforms as T

            transforms = T.Compose(
                [
                    T.ToTensor(),
                    T.CenterCrop(300),
                    T.Resize((244, 244)),
                    T.RandomVerticalFlip(p=1.0),
                    T.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)),
                ]
            )

            crafted_img = transforms(image_batch[0]).cpu().numpy()
            crafted_img = np.transpose(crafted_img, (1, 2, 0))


    :param transforms: List of transforms that are applied sequentially
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(
        self,
        transforms: List[Union[str, Dict[str, Dict]]] = [
            {'Resize': dict(size=(224, 224))},
            {'Normalize': dict(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))},
        ],
        *args,
        **kwargs,
    ):
        """Set constructor."""
        super().__init__(*args, **kwargs)
        if not isinstance(transforms, list):
            self.logger.error('The `transforms` argument has to be a list.')
            raise ValueError

        self.transforms_specification = transforms

    def post_init(self):
        from torchvision import transforms as T

        transforms_list = []
        for transform in self.transforms_specification:
            if isinstance(transform, str):
                tr_name, tr_kwargs = transform, {}
            elif isinstance(transform, dict):
                tr_name, tr_kwargs = next(iter(transform.items()))
            else:
                self.logger.error(
                    'Each element of the `transform` has to be either a dict or  a string'
                )
                raise ValueError

            try:
                tr_class = getattr(T, tr_name)
            except AttributeError:
                self.logger.error(
                    f'The transform class `{tr_name}` does not seem to exist,'
                    'check that you have not made a typo.'
                )
                raise ValueError

            try:
                transform_instance = tr_class(**tr_kwargs)
            except:
                self.logger.error(
                    f'Error instantiating torchvision transforms class `{tr_name}`,'
                    ' check if all kwargs are valid and if transform is valid for tensors.'
                )
                raise ValueError

            # Removes randomness in transforms
            # does not handle inherently random transform like `RandomErasing`
            # where p is probabilty of being random
            transform_instance.__dict__.update({'p': 1.0})
            transforms_list.append(transform_instance)
        self.transforms = T.Compose(transforms_list)

    def craft(
        self, blob: Union[List['np.ndarray'], 'np.ndarray'], *args, **kwargs
    ) -> List[Dict]:
        """Apply the transforms to a batch of numpy images
        Note: The numpy arrays are standardize to range [0, 1] before applying transforms

        :param blob: numpy images in formats `(..., H, W, C)`, `(H, W, C)`, `list[(H, W, C)]`
            with value [0, 255] range
        :return: list of dictionaries, each containing a transformed image
        """
        if isinstance(blob, list):
            blob = np.stack(blob)
        elif isinstance(blob, np.ndarray):
            if blob.ndim == 3:
                blob = np.stack([blob])
        else:
            self.logger.error(f'Expected numpy array or list of numpy arrays')
            raise ValueError
        assert (
            blob.ndim == 4
        ), 'numpy should be of shape (..., H, W, C) or list[(H, W, C)]'

        blob_ = self._to_tensor(blob)
        blob_ = self.transforms(blob_)
        blob_ = self._to_numpy(blob_)

        dict_list = []
        for sample in blob_:
            dict_list.append({'blob': sample})
        return dict_list

    @staticmethod
    def _to_tensor(array: 'np.ndarray') -> 'torch.Tensor':
        """Convert numpy to tensor
        Replaces `transforms.ToTensor` to allow batching

        :param array: numpy array of format `[B, H, W, C]`
        :return: standardized tensor in the format `[B, C, H, W]` as required by torchvision transforms
        """
        import torch

        array = np.transpose(array, (0, 3, 1, 2))
        if not array.data.contiguous:
            array = np.ascontiguousarray(array)
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        return torch.from_numpy(array).float().to(device).div(255)

    @staticmethod
    def _to_numpy(tensor: 'torch.Tensor') -> 'np.ndarray':
        """Convert tensor to numpy

        :param tensor: tensor in the format `[B, C, H, W]`
        :return: numpy array of format `[B, H, W, C]`
        """

        array = tensor.detach().cpu().numpy()
        array = np.transpose(array, (0, 2, 3, 1))
        return array



import io
from typing import Dict, List

import fitz
import numpy as np
import pdfplumber
from jina.executors.decorators import single
from jina.executors.segmenters import BaseSegmenter
from pdftitle import get_title_from_file, get_title_from_io


class PDFPlumberSegmenter(BaseSegmenter):
    """
    :class:`PDFPlumberSegmenter` Extracts data (text and images) from PDF files.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _extract_img(self, pdf_img, chunks):
        with pdf_img:
            for page in range(len(pdf_img)):
                for img in pdf_img.getPageImageList(page):
                    xref = img[0]
                    pix = fitz.Pixmap(pdf_img, xref)
                    # read data from buffer and reshape the array into 3-d format
                    np_arr = np.frombuffer(pix.samples, dtype=np.uint8).reshape(pix.h, pix.w, pix.n).astype('float32')
                    if pix.n - pix.alpha < 4:  # if gray or RGB
                        if pix.n == 1:  # convert gray to rgb
                            np_arr_rgb = np.concatenate((np_arr,) * 3, -1)
                            chunks.append(dict(blob=np_arr_rgb, weight=1.0, mime_type='image/png'))
                        elif pix.n == 4:  # remove transparency layer
                            np_arr_rgb = np_arr[..., :3]
                            chunks.append(dict(blob=np_arr_rgb, weight=1.0, mime_type='image/png'))
                        else:
                            chunks.append(dict(blob=np_arr, weight=1.0, mime_type='image/png'))
                    else:  # if CMYK:
                        pix = fitz.Pixmap(fitz.csRGB, pix)  # Convert to RGB
                        np_arr = np.frombuffer(pix.samples, dtype=np.uint8).reshape(pix.h, pix.w, pix.n).astype(
                            'float32')
                        chunks.append(
                            dict(blob=np_arr, weight=1.0, mime_type='image/png'))

    def _extract_text(self, pdf_content, chunks, title):
        with pdf_content:
            count = len(pdf_content.pages)
            for i in range(count):
                page = pdf_content.pages[i]
                text_page = page.extract_text(x_tolerance=1, y_tolerance=1)
                text_page = text_page.replace(u'\xa0', u'')
                if title:
                    chunks.append(dict(text=title, weight=1.0, mime_type='text/plain', tags={'title': True}))
                if text_page:
                    chunks.append(dict(text=text_page, weight=1.0, mime_type='text/plain'))

    @single(slice_nargs=3)
    def segment(self, uri: str, buffer: bytes, mime_type: str, *args, **kwargs) -> List[Dict]:
        """
        Segements PDF files. Extracts data from them.

        Checks if the input is a string of the filename,
        or if it's the file in bytes.
        It will then extract the data from the file, creating a list for images,
        and text.

        :param uri: File name of PDF
        :type uri: str
        :param buffer: PDF file in bytes
        :type buffer: bytes
        :param mime_type: the type of data
        :returns: A list of documents with the extracted data
        :rtype: List[Dict]
        """
        chunks = []
        if mime_type != 'application/pdf':
            return chunks

        if uri:
            try:
                pdf_img = fitz.open(str(uri))
                pdf_content = pdfplumber.open(uri)
                title = get_title_from_file(uri)
            except Exception as ex:
                self.logger.error(f'Failed to open {uri}: {ex}')
                return chunks
        elif buffer:
            try:
                pdf_img = fitz.open(stream=buffer, filetype='pdf')
                pdf_content = pdfplumber.open(io.BytesIO(buffer), password=b"")
                title = get_title_from_io(io.BytesIO(buffer))
            except Exception as ex:
                self.logger.error(f'Failed to load from buffer')
                return chunks
        else:
            self.logger.warning('No value found in `buffer` or `uri`')
            return chunks
        # Extract images
        self._extract_img(pdf_img, chunks)
        # Extract text
        self._extract_text(pdf_content, chunks, title)
        return chunks



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
import itertools
import re
from typing import Callable, List, Tuple, Optional, Iterable
from collections import Counter


class BKTree:
    """
    :class:`BKTree` allows fast search of words up to n distances away

    :param distfn : Function to compute distances between words, like edit distance function.
    :param words : iterable of words to build the tree.
    :param sort_candidates : If true results will be sorted by distance.
    """

    def __init__(self, distfn: Callable, words: List[str], sort_candidates: bool = False):
        self.distfn = distfn
        self.sort_candidates = sort_candidates

        it = iter(words)
        root = next(it)
        self.tree = (root, {})

        for i in it:
            self._add_word(self.tree, i)

    def _add_word(self, parent, word):
        pword, children = parent
        d = self.distfn(word, pword)
        if d in children:
            self._add_word(children[d], word)
        else:
            children[d] = (word, {})

    def _search_descendants(self, parent, n, distance, query_word):
        """
        :param parent : parent in the tree
        :param n : maximum edit distance allowed.
        :param distance : function used to compute the edit distance.
        :param query_word : query word
        """
        node_word, children_dict = parent
        dist_to_node = distance(query_word, node_word)
        self.visited_nodes.append(node_word)
        results = []
        if dist_to_node <= n:
            results.append((dist_to_node, node_word))

        for i in range(dist_to_node - n, dist_to_node + n + 1):
            child = children_dict.get(i)
            if child is not None:
                results.extend(self._search_descendants(child, n, distance, query_word))

        return results

    def query(self, query_word: str, n: int, return_distances: bool = False):
        """
        :param query_word : anchor word used to find candidates.
        :param n : maximum edit distance allowed
        :param return_distances : whether to return distances (True means return distance)
        """
        self.visited_nodes = []

        distance_candidate_list = self._search_descendants(self.tree, n, self.distfn, query_word)
        if self.sort_candidates:
            distance_candidate_list = sorted(distance_candidate_list)

        if return_distances:
            return distance_candidate_list
        else:
            return [x[1] for x in distance_candidate_list]


class PyNgramSpell:
    """
    :class:`PyNgramSpell` Corrects misspelled words from strings.

    :param ngram_range : ngram_range (NOW THIS IS FIXED future versions will improve).
    :param tokenizer : tokenizer function to tokenize an input string.
    :param string_preprocessor_func : function used to pre-process the string, `lower` by default.
    :param token_pattern : Regular expression used to tokenize an input string.
    :param lambda_interpolation : Regularization assigned to the interpolation.
    :param min_freq : Minimum frequency for a word to be  in the vocabulary.
    :param max_dist : Maximum edit distance allowed to generate candidate tokens.
    :param sort_candidates : Provide candidates sorted by edit distance.
    :param bktree : True if bktree is used to search candidates. False means exhaustive search.
    """

    def __init__(self,
                 ngram_range: Tuple[int] = (1, 2),
                 token_pattern: str = r"(?u)\b\w\w+\b",
                 lambda_interpolation: float = 0.3,
                 min_freq: int = 5,
                 max_dist: int = 1,
                 sort_candidates: bool = False,
                 use_bktree: bool = True,
                 tokenizer: Optional[Callable] = None,
                 string_preprocessor_func: Callable = str.lower
                 ):

        self.ngram_range = ngram_range
        self.string_preprocessor_func = string_preprocessor_func
        self.token_pattern = token_pattern
        self.lambda_interpolation = lambda_interpolation
        self.min_freq = min_freq
        self.max_dist = max_dist
        self.sort_candidates = sort_candidates
        self.use_bktree = use_bktree
        self.tokenizer = tokenizer
        self.vocabulary = None
        self.bktree = None

    def _build_tokenizer(self):
        """Return a function that splits a string into a sequence of tokens.
        """
        if self.tokenizer is not None:
            return self.tokenizer
        token_pattern = re.compile(self.token_pattern)

        if token_pattern.groups > 1:
            raise ValueError(
                "More than 1 capturing group in token pattern. Only a single "
                "group should be captured."
            )

        return token_pattern.findall

    def _get_candidates_bktree(self, token: str, max_dist: int):
        """Return a list of candidate words from the vocabulary at most `max_dist` away from the input token.
        :param token : token to be corrected.
        :param max_dist : maximum allowed edit distance.
        """
        candidate_tokens = self.bktree.query(token, max_dist)

        if len(candidate_tokens) > 0:
            return candidate_tokens
        else:
            return [token]

    def _get_candidates_exhaustive(self, token: str, max_dist: int):
        """Return a list of candidate words from the vocabulary at most `max_dist` away from the input token.
        This version of the function is private and kept for benchmarking pourposes. This function computes the
        edit distance between the input token and all words in the vocabulary. Then it filters candidates by
        the edit distance.
        :param token : word to be corrected.
        :param max_dist : maximum distance allowed for candidates.
        """
        from editdistance import eval as edit_distance

        token = token.lower()
        distance_token_to_words = {word: edit_distance(word, token) for word in self.vocabulary}
        min_dist = min(distance_token_to_words.values())
        if min_dist <= max_dist:

            if self.sort_candidates:
                result = sorted(
                    [(distance, word) for word, distance in distance_token_to_words.items() if distance <= max_dist])
            else:
                result = [word for word, distance in distance_token_to_words.items() if distance <= max_dist]

            return result
        else:
            return [token]

    def _get_candidates(self, token: str, max_dist: int):
        """Get the candidate words with at most `max_dist`
        :param token : token to be corrected.
        :param max_dist : maximum allowed edit distance.
        """
        if self.bktree is not None:
            return self._get_candidates_bktree(token, max_dist)
        else:
            return self._get_candidates_exhaustive(token, max_dist)

    def _filter_vocabulary(self, min_freq: int) -> None:
        """Remove words from the vocabulary that do not have more than `min_freq`
        :param min_freq : minimum frequency required to be in the vocabulary
        """
        self.vocabulary = set(dict(filter(lambda x: x[1] > min_freq, self.unigram_freq_dict.items())).keys())

    def _correct_with_bigrams(self, tokenized_sentence: List[int]) -> List[int]:
        """Correct the words in the tokenized_sentence that are not part of the vocabulary
        :param tokenized_sentence : sentence tokenized
        """

        def _prob_word(word):
            return self.unigram_freq_dict.get(word, 0) / len(self.unigram_freq_dict)

        def _probability_bigram(word1, word2):
            bigram = ((word1, word2))
            if self.bigram_freq_dict.get(bigram, 0) == 0:
                return 0
            else:
                return self.bigram_freq_dict.get(bigram, 0) / self.unigram_freq_dict[word1]

        def _interpolation_probability(word1, word2):
            return (1 - self.lambda_interpolation) * _probability_bigram(word1, word2) + \
                   self.lambda_interpolation * _prob_word(word2)

        previous_word = '.'
        for index, word in enumerate(tokenized_sentence):
            if word not in self.vocabulary:
                candidates = {candidate: _interpolation_probability(previous_word, candidate) for candidate in
                              self._get_candidates(word, max_dist=self.max_dist)}

                tokenized_sentence[index] = max(candidates, key=candidates.get)
            previous_word = tokenized_sentence[index]

        return tokenized_sentence

    def fit(self, X: Iterable[str]):
        """Fit the ngram model and the vocabulary from the training data.
        :param X : Iterable over strings containing the corpus used to train the spellcheker.
        """
        from nltk.collocations import BigramCollocationFinder
        from editdistance import eval as edit_distance

        self.tokenize_func = self._build_tokenizer()
        X_tokenized = [self.tokenize_func(self.string_preprocessor_func(x)) for x in X]
        self.unigram_freq_dict = dict(Counter(itertools.chain(*X_tokenized)))
        bigram_finder = BigramCollocationFinder.from_documents(X_tokenized)
        self.bigram_freq_dict = dict(bigram_finder.ngram_fd.items())
        self.vocabulary = set(list(itertools.chain(*self.bigram_freq_dict.keys())))

        if self.min_freq > 0:
            self._filter_vocabulary(min_freq=self.min_freq)

        if self.use_bktree:
            self.bktree = BKTree(edit_distance, self.vocabulary, sort_candidates=self.sort_candidates)

    def transform(self, x: str, tokenize: bool = True) -> str:
        """Corrects the misspelled words on each element from X.
        :param x : string to be corrected
        :param tokenize:

        """
        x = self.string_preprocessor_func(x)

        if tokenize:
            tokenized_sentence = self.tokenize_func(x)
        else:
            tokenized_sentence = x.copy()

        tokenized_sentence = self._correct_with_bigrams(tokenized_sentence)

        return tokenized_sentence

    def save(self, file_path: str):
        """Save the instance an PyNgramSpell to the provided path.
        :param file_path : path where to store the spell checker object
        """
        import pickle
        with open(file_path, 'wb') as f:
            pickle.dump(self, f)



from .document import DocumentSet
from ...helper import deprecated_class

from ..arrays.match import MatchArray


@deprecated_class(new_class=MatchArray)
class MatchSet(DocumentSet):
    """
    :class:`MatchSet` is deprecated. A new class name is MatchArray.
    """

    pass



import os
import sys

from pkg_resources import resource_filename

from jina.parsers.base import set_base_parser
from jina.parsers.helper import add_arg_group
from jina.parsers.peapods.base import mixin_base_ppr_parser
from jina.parsers.peapods.runtimes.remote import mixin_remote_parser


def mixin_daemon_parser(parser):
    gp = add_arg_group(parser, title='Daemon')

    gp.add_argument(
        '--no-fluentd',
        action='store_true',
        default=False,
        help='do not start fluentd, no log streaming',
    )

    gp.add_argument(
        '--workspace',
        type=str,
        default='/tmp/jinad',
        help='the directory for storing all uploaded dependencies',
    )


def get_main_parser():
    parser = set_base_parser()

    mixin_remote_parser(parser)
    mixin_base_ppr_parser(parser)
    mixin_daemon_parser(parser)

    parser.set_defaults(
        port_expose=8000,
        log_config=os.getenv(
            'JINAD_LOG_CONFIG',
            resource_filename('jina', '/'.join(('resources', 'logging.daemon.yml'))),
        ),
    )

    return parser


def _get_run_args(print_args: bool = True):
    from jina.helper import colored
    from . import daemon_logger

    parser = get_main_parser()
    from argparse import _StoreAction, _StoreTrueAction

    args = parser.parse_args()
    if print_args:
        from pkg_resources import resource_filename

        default_args = {
            a.dest: a.default
            for a in parser._actions
            if isinstance(a, (_StoreAction, _StoreTrueAction))
        }

        with open(
            resource_filename('jina', '/'.join(('resources', 'jina.logo')))
        ) as fp:
            logo_str = fp.read()
        param_str = []
        for k, v in sorted(vars(args).items()):
            j = f'{k.replace("_", "-"): >30.30} = {str(v):30.30}'
            if default_args.get(k, None) == v:
                param_str.append('   ' + j)
            else:
                param_str.append('🔧️ ' + colored(j, 'blue', 'on_yellow'))
        param_str = '\n'.join(param_str)
        daemon_logger.info(f'\n{logo_str}\n▶️  {" ".join(sys.argv)}\n{param_str}\n')
    return args



import os

from pathlib import Path
from pyngramspell import PyNgramSpell

cur_dir = os.path.dirname(os.path.abspath(__file__))


def get_evaluation(n_test=100):
    lines = open(os.path.join(cur_dir, 'evaluation.txt')).readlines()
    data = []

    import nltk

    for line in lines:
        indexes = []
        tokenize_line = nltk.word_tokenize(line)
        misspelled_line = tokenize_line.copy()
        corrected_line = tokenize_line.copy()
        for index, word in enumerate(tokenize_line):
            if '|' in word:
                misspelled_line[index] = word.split('|')[0]
                corrected_line[index] = word.split('|')[1]
                indexes.append(index)
        data.append({'original': misspelled_line, 'corrected': corrected_line, 'indexes': indexes})

    test = data[:n_test]
    train = data[n_test:]
    return train, test


def train_model(model_path: str = os.path.join(cur_dir, '../model/model.pickle')):
    import nltk
    nltk.download('punkt')
    train, test = get_evaluation()
    y_tr = [x['corrected'] for x in train]
    train_data = [' '.join(x) for x in y_tr]
    speller = PyNgramSpell()
    speller.fit(train_data)
    assert ' '.join(speller.transform('the maan went to the store')) == 'the man went to the store'
    os.makedirs(Path(model_path).parent, exist_ok=True)
    speller.save(model_path)


if __name__ == '__main__':
    train_model()



from .document import DocumentSet
from ...helper import deprecated_class
from ..arrays.chunk import ChunkArray


@deprecated_class(new_class=ChunkArray)
class ChunkSet(DocumentSet):
    """
    :class:`ChunkSet` is deprecated. A new class name is ChunkArray.
    """

    pass



from collections.abc import MutableSequence
from typing import Iterable, Union, Dict

from ...helper import deprecated_class

from ..querylang import QueryLang
from ...proto.jina_pb2 import QueryLangProto

from ..arrays.querylang import QueryLangArray

AcceptQueryLangType = Union[QueryLang, QueryLangProto, Dict]

__all__ = ['QueryLangSet', 'AcceptQueryLangType']


@deprecated_class(new_class=QueryLangArray)
class QueryLangSet(MutableSequence):
    """
    :class:`QueryLangSet` is deprecated. A new class name is QueryLangArray.
    """

    pass



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
from typing import Optional

from jina.executors.decorators import single
from jina.executors.crafters import BaseCrafter


class PySpellChecker(BaseCrafter):

    """
    :class:`PySpellChecker` wraps pyspellchecker (https://github.com/barrust/pyspellchecker) library
    to provide spelling correction capacity as a crafter in Jina

    :param language: The language of the dictionary to load or None \
            for no dictionary. Supported languages are `en`, `es`, `de`, `fr`, \
            `pt` and `ru`. Defaults to `en`. A list of languages may be \
            provided and all languages will be loaded.
    :param local_dictionary: he path to a locally stored word \
            frequency dictionary; if provided, no language will be loaded
    :param distance: The edit distance to use. Defaults to 2.
    :param case_sensitive: Flag to use a case sensitive dictionary or \
            not, only available when not using a language dictionary.
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments

    """

    def __init__(self,
                 language: str = 'en',
                 local_dictionary: Optional[str] = None,
                 distance: int = 2,
                 case_sensitive: bool = False,
                 *args, **kwargs):
        """Set constructor."""
        super().__init__(*args, **kwargs)
        self.language = language
        self.local_dictionary = local_dictionary
        self.distance = distance
        self.case_sensitive = case_sensitive

    def post_init(self):
        from spellchecker import SpellChecker

        super().post_init()
        self.speller = SpellChecker(language=self.language,
                                    local_dictionary=self.local_dictionary,
                                    distance=self.distance,
                                    case_sensitive=self.case_sensitive)

    @single
    def craft(self, text: str, *args, **kwargs):
        """
        Craft sentences correcting misspelled words

        :param text: The text to be corrected
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: A dictionary with the extracted text
        """
        words = self.speller.split_words(text)
        corrected_text = ' '.join([self.speller.correction(word) for word in words])
        return dict(text=corrected_text)



from .traversable import TraversableSequence
from ..arrays.doc_groundtruth import DocumentGroundtruthSequence
from ...helper import deprecated_class


@deprecated_class(
    new_class=DocumentGroundtruthSequence,
    custom_msg="The class has been moved to '..types.arrays', keeping its original name.",
)
class DocumentGroundtruthSequence(TraversableSequence):
    """
    :class:`DocumentGroundtruthSequence` is deprecated. It moved to `jina.types.array.doc_groundtruth`.
    """

    pass



from collections.abc import MutableSequence

from ...helper import deprecated_class

from .traversable import TraversableSequence

from ..arrays.document import DocumentArray

__all__ = ['DocumentSet']


@deprecated_class(new_class=DocumentArray)
class DocumentSet(TraversableSequence, MutableSequence):
    """
    :class:`DocumentSet` is deprecated. A new class name is ChunkArray.
    """

    pass



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os

from jina.executors.decorators import single
from jina.executors.crafters import BaseCrafter
from jina.excepts import PretrainedModelFileDoesNotExist

cur_dir = os.path.dirname(os.path.abspath(__file__))


class SpellChecker(BaseCrafter):

    def __init__(self,
                 model_path: str = os.path.join(cur_dir, 'model/model.pickle'),
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_path = model_path

    def post_init(self):
        super().post_init()
        import pickle
        from .pyngramspell import PyNgramSpell, BKTree

        self.model = None
        if os.path.exists(self.model_path):
            with open(self.model_path, 'rb') as model_file:
                self.model = pickle.load(model_file)
        else:
            raise PretrainedModelFileDoesNotExist(
                f'{self.model_path} not found, cannot find a fitted spell checker'
            )

    @single
    def craft(self, text: str, *args, **kwargs):
        return dict(text=' '.join(self.model.transform(text)))



"""Module for Jina Requests."""
__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterator, Union, Tuple, AsyncIterable, Iterable, Optional

from .helper import _new_request_from_batch
from ... import Request
from ...enums import RequestType, DataInputType
from ...helper import batch_iterator
from ...logging import default_logger
from ...types.document import DocumentSourceType, DocumentContentType, Document
from ...types.arrays.querylang import AcceptQueryLangType

SingletonDataType = Union[
    DocumentContentType,
    DocumentSourceType,
    Document,
    Tuple[DocumentContentType, DocumentContentType],
    Tuple[DocumentSourceType, DocumentSourceType],
]

GeneratorSourceType = Union[
    Document, Iterable[SingletonDataType], AsyncIterable[SingletonDataType]
]


def request_generator(
    data: GeneratorSourceType,
    request_size: int = 0,
    mode: RequestType = RequestType.INDEX,
    mime_type: Optional[str] = None,
    queryset: Optional[
        Union[AcceptQueryLangType, Iterator[AcceptQueryLangType]]
    ] = None,
    data_type: DataInputType = DataInputType.AUTO,
    **kwargs,  # do not remove this, add on purpose to suppress unknown kwargs
) -> Iterator['Request']:
    """Generate a request iterator.

    :param data: the data to use in the request
    :param request_size: the request size for the client
    :param mode: the request mode (index, search etc.)
    :param mime_type: mime type
    :param queryset: querylang set of queries
    :param data_type: if ``data`` is an iterator over self-contained document, i.e. :class:`DocumentSourceType`;
            or an iterator over possible Document content (set to text, blob and buffer).
    :param kwargs: additional arguments
    :yield: request
    """

    _kwargs = dict(mime_type=mime_type, weight=1.0, extra_kwargs=kwargs)

    try:
        if not isinstance(data, Iterable):
            data = [data]
        for batch in batch_iterator(data, request_size):
            yield _new_request_from_batch(
                _kwargs, batch, data_type, mode, queryset, **kwargs
            )

    except Exception as ex:
        # must be handled here, as grpc channel wont handle Python exception
        default_logger.critical(f'inputs is not valid! {ex!r}', exc_info=True)
